import SimpleITK as sitk
import numpy as np
from loguru import logger

from mircat_stats.statistics.nifti import MircatNifti, _resample
from mircat_stats.statistics.utils import _filter_largest_components
from mircat_stats.configs.models import torch_model_configs


class SegNotFoundError(ValueError):
    """
    Raised when the aorta segmentation is not found
    """

    pass


class SegmentationSizeDoesNotMatchError(ValueError):
    """
    Raised when the segmentation and CT image do not have the same size
    """

    pass


class Segmentation:
    """Class to filter one or multiple segmentations out from a single model and
    hold them in a single object. This is useful for specific morphology-based statistics
    """

    def __init__(self, nifti: MircatNifti, seg_names: list[str], reload_gaussian: bool = False):
        """Initialize Segmentation class.

        This class handles filtering and potentially analysis of segmented CT images.
        It will load and filter the appropriate complete segmentation on initialization.

        Args:
            nifti (MircatNifti): A MircatNifti object containing CT and segmentation data
            seg_names (list[str]): List of segmentation names to analyze

        Attributes:
            original_ct: Original CT image data
            vert_midlines: Vertebrae midline data
            seg_folder: Folder containing segmentation files
            seg_info: Dictionary containing segmentation information
            model: Model used for segmentation
            segmentation: Filtered segmentation image
            seg_names: List of segmentation names in output
        """
        self.original_ct = nifti.original_ct
        self.vert_midlines = nifti.vert_midlines
        self.seg_folder = nifti.seg_folder
        self.seg_names = seg_names
        self._find_seg_model()
        self._filter_to_segmentation(nifti, reload_gaussian)

    def _find_seg_model(self):
        seg_info = {}
        for seg_name in self.seg_names:
            for model in torch_model_configs:
                if seg_name in torch_model_configs[model]["output_map"]:
                    seg_model = model
                    seg_idx = torch_model_configs[model]["output_map"][seg_name]
                    break
            seg_info[seg_name] = {"model": seg_model, "idx": seg_idx}
        model = set(info["model"] for info in seg_info.values())
        if len(model) > 1:
            raise ValueError("All segmentations must come from the same model")
        model = model.pop()
        self.seg_info = seg_info
        self.model = model

    def _filter_to_segmentation(self, nifti: MircatNifti, reload_gaussian: bool) -> sitk.Image:
        """Filter input nifti to segmented regions.

        This method applies filtering to convert a nifti image into segmented regions.
        Labels will be indexed from 1 to len(labels).

        Args:
            nifti (MircatNifti): Input nifti image to be segmented.
            reload_gaussian (bool): Whether to reload the segmentation using gaussian smoothing. Useful if you loaded the segmentation without smoothing.

        Returns:
            sitk.Image: Filtered image containing segmented regions.
        """
        if self.model == "total":
            complete = nifti.total_seg
            seg_file = nifti.seg_files["total"]
        elif self.model == "body":
            complete = nifti.body_seg
            seg_file = nifti.seg_files["body"]
        elif self.model == "tissues":
            complete = nifti.tissues_seg
            seg_file = nifti.seg_files["tissues"]
        if reload_gaussian:
            complete = sitk.ReadImage(seg_file)
        labels = list(self.seg_info.keys())
        label_indices = [v["idx"] for v in self.seg_info.values()]

        label_map = {old_idx: new_idx for new_idx, old_idx in enumerate(label_indices, start=1)}
        seg_arr = sitk.GetArrayFromImage(complete).astype(np.uint8)
        mask = np.isin(seg_arr, label_indices)
        seg_arr[~mask] = 0
        for old_idx, new_idx in label_map.items():
            seg_arr[seg_arr == old_idx] = new_idx
        mapped_indices = [int(x) for x in np.unique(seg_arr) if x != 0]
        if 1 not in mapped_indices:
            logger.opt(exception=True).error("No segmentations found in the input")
            raise SegNotFoundError("No segmentations found in the input")
        if set(mapped_indices) != set(label_map.values()):
            missing = set(label_map.values()).difference(set(mapped_indices))
            missing_labels = ",".join([labels[idx - 1] for idx in missing])
            logger.debug(f"{missing_labels} not found in the input")
            labels = [labels[idx - 1] for idx in mapped_indices]
        segmentation = sitk.GetImageFromArray(seg_arr)
        segmentation.CopyInformation(complete)
        if reload_gaussian:
            # Get cleaner segmentation using gaussian smoothing
            segmentation = _resample(segmentation, (1.0, 1.0, 1.0), is_label=True, gaussian=True)
        self.segmentation = _filter_largest_components(segmentation, mapped_indices)
        self.seg_names = labels

    def extract_segmentation_bounding_box(self, padding: tuple[int] | int = (0, 0, 0)) -> tuple[sitk.Image, sitk.Image]:
        """
        Extract the bounding box of the segmentation with a given amount of padding around it.
        Args
        ----
        padding: tuple[int] | int
            The padding to add around the bounding box. If a single integer is given, the same padding will be added in all directions.
        Returns
        -------
        tuple[sitk.Image, sitk.Image]
            The cropped segmentation and CT image
        """
        assert self.segmentation.GetSize() == self.original_ct.GetSize(), SegmentationSizeDoesNotMatchError(
            "Segmentation and CT image must have the same size"
        )
        if isinstance(padding, int):
            padding = (padding, padding, padding)
        else:
            assert len(padding) == 3, ValueError(
                "Bounding box padding must be a single integer or a tuple of 3 integers"
            )

        # Set up sitk filter
        bbox_filter = sitk.LabelShapeStatisticsImageFilter()
        bbox_filter.SetComputeOrientedBoundingBox(True)
        bbox_filter.Execute(self.segmentation)
        bbox = bbox_filter.GetBoundingBox(1)

        # Set up the cropping filter
        start_idx = list(bbox[0:3])
        size = list(bbox[3:6])
        for i in range(3):
            # Adjust start index
            start_idx[i] = max(0, start_idx[i] - padding[i])
            # Adjust size to account for padding and image bounds
            max_size = self.segmentation.GetSize()[i] - start_idx[i]
            size[i] = min(size[i] + 2 * padding[i], max_size)
        # Extract regions using the same coordinates for both images
        extract = sitk.ExtractImageFilter()
        extract.SetSize(size)
        extract.SetIndex(start_idx)

        # Extract from segmentation
        cropped_seg = extract.Execute(self.segmentation)
        cropped_img = extract.Execute(self.original_ct)
        return cropped_seg, cropped_img


class Vessel(Segmentation):
    """Child class of Segmentation to filter one or multiple vessel segmentations out from a single model and
    hold them in a single object. This is useful for specific morphology-based statistics. Has specific implementations of centerline and CPR generation.
    """

    def __init__(self, nifti: MircatNifti, seg_names: list[str]):
        """Initialize Vessel class.

        This class handles filtering and potentially analysis of segmented CT images.
        It will load and filter the appropriate complete segmentation on initialization.

        Args:
            nifti (MircatNifti): A MircatNifti object containing CT and segmentation data
            seg_names (list[str]): List of segmentation names to analyze

        Attributes:
            original_ct: Original CT image data
            vert_midlines: Vertebrae midline data
            seg_folder: Folder containing segmentation files
            seg_info: Dictionary containing segmentation information
            model: Model used for segmentation
            segmentation: Filtered segmentation image
            seg_names: List of segmentation names in output
        """
        super().__init__(nifti, seg_names)



import argparse
from loguru import logger
from pathlib import Path
from mircat_stats.configs.logging import configure_logging, get_project_root
from threadpoolctl import threadpool_limits
from mircat_stats.dicom import convert_dicom_folders_to_nifti, update
from mircat_stats.statistics import calculate_nifti_stats
import shutil


def mircat_stats():
    """
    MirCAT Stats Only CLI tool
    """
    parser = argparse.ArgumentParser(description="Mircato Stats Only CLI tool")
    parser.add_argument("-q", "--quiet", help="Decrease output verbosity", action="store_true")
    subparsers = parser.add_subparsers(dest="command", required=True)

    # Create the 'convert' subparser
    convert_parser = subparsers.add_parser("convert", help="Convert DICOM files")
    convert_parser.add_argument("dicoms", help="Path to DICOM files", type=Path)
    convert_parser.add_argument("output_dir", help="Output directory", type=Path)
    convert_parser.add_argument("-n", "--num-workers", help="Number of something", type=int, default=1)
    convert_parser.add_argument(
        "-ax",
        "--axial-only",
        help="Only convert axial dicom series",
        action="store_true",
    )
    convert_parser.add_argument("-nm", "--no-mip", help="Do not convert likely mip series", action="store_true")
    convert_parser.add_argument(
        "-th",
        "--threads",
        help="Number of threads for each worker",
        type=int,
        default=1,
    )
    # Set up stats parser
    stats_parser = subparsers.add_parser("stats", description="Calculate statistics for NIfTI files")
    stats_parser.add_argument(
        "niftis",
        type=Path,
        help="NIfTI file or a text file containing paths to multiple NIfTI files to calculate statistics for",
    )
    stats_parser.add_argument(
        "-t",
        "--task-list",
        type=str,
        nargs="+",
        default=["total", "contrast", "aorta", "tissues"],
        help='List of statistics tasks to perform. Default = ["total", "contrast", "aorta", "tissues"]',
    )
    stats_parser.add_argument(
        "-n",
        "--num-workers",
        type=int,
        default=4,
        help="Number of workers to use for multi-process operations. Default=1",
    )
    stats_parser.add_argument(
        "-th",
        "--threads",
        type=int,
        default=4,
        help="Number of threads each worker will use. Default=4",
    )
    stats_parser.add_argument(
        "-mc",
        "--mark-complete",
        action="store_true",
        help="Mark the statistics as complete regardless of stats performed",
    )
    stats_parser.add_argument(
        "-g",
        "--gaussian",
        action="store_true",
        help="Apply a gaussian smoothing to the label segmentations. Will be slower but more precise upon scaling",
    )
    # Create update parser
    update_parser = subparsers.add_parser(
        "update",
        help="Update the header and stats data for a NIfTI file to the latest version",
    )
    update_parser.add_argument("niftis", help="Path to NIfTI files", type=Path)
    update_parser.add_argument("-n", "--num-workers", help="Number of workers", type=int, default=1)
    update_parser.add_argument("-th", "--threads", help="Number of threads", type=int, default=1)

    args = parser.parse_args()
    args.verbose = not args.quiet
    threadpool_limits(limits=args.threads)

    if args.command == "convert":
        if args.dicoms.is_dir():
            logfile = "./nifti_conversion_log.json"
            dicom_list = [str(args.dicoms)]
        else:
            logfile = f'{args.dicoms.with_suffix("")}_conversion_log.json'
            with args.dicoms.open() as f:
                dicom_list = f.read().splitlines()
        configure_logging(logger, logfile, args.verbose)
        convert_dicom_folders_to_nifti(
            dicom_list,
            args.output_dir,
            args.num_workers,
            args.axial_only,
            args.no_mip,
            args.verbose,
        )

    elif args.command == "stats" or args.command == "update":
        # If the input to the niftis argument is just a singular nifti file, make it a list and log in the same dir
        if args.niftis.suffixes == [".nii", ".gz"] or args.niftis.suffix == ".nii":
            logfile = str(args.niftis).replace(".nii", "").replace(".gz", "")
            logfile = f"{logfile}_{args.command}_log.jsonl"
            nifti_list = [str(args.niftis)]
        # Otherwise, make the log in the same dir as the input file and read the list
        else:
            logfile = f"{args.niftis.with_suffix('')}_{args.command}_log.jsonl"
            with args.niftis.open() as f:
                nifti_list = [x for x in f.read().splitlines()]
        configure_logging(logger, logfile, args.verbose)

        if args.command == "stats":
            calculate_nifti_stats(
                nifti_list,
                args.task_list,
                args.num_workers,
                args.threads,
                args.mark_complete,
                args.gaussian,
            )
        elif args.command == "update":
            update(nifti_list, args.num_workers)
        else:
            print("Unknown command")



# Module for curved planar reformation from centerline. Only good for straightened CPR currently
import numpy as np
from loguru import logger
from scipy.interpolate import CubicSpline, interpn
from skimage import measure
from skimage.measure._regionprops import RegionProperties
from skimage.filters import gaussian
from skimage.morphology import remove_small_holes


###### CPR Generation Functions ###########
def create_straightened_cpr(
    vessel: np.ndarray,
    centerline: np.ndarray,
    cross_section_xy: tuple,
    resolution: int = 1,
    sigma: int = 2,
    is_binary: bool = True,
) -> np.ndarray:
    """Create a straightened Curved Planar Reformation of a vessel using a centerline
    :param vessel: the vessel as a numpy array
    :param centerline: the centerline as a numpy array
    :param cross_section_xy: the number of points in (x, y) for the cross-section;
        i.e. (100, 100) will result in an extracted cross-section of shape (100, 100)
    :param resolution: the resolution to sample the cpr. Default = 1
    :param sigma: the sigma value for the gaussian filter. Default = 2
    :return: a numpy array containing the straightened curved planar reformation
    """
    tangents = _compute_tangent_vectors(centerline)
    cpr = []
    num_touching = 0
    for center_point, normal_vector in zip(centerline, tangents):
        cross_section = _extract_cross_sectional_slice(
            vessel, center_point, normal_vector, cross_section_xy, resolution, is_binary
        )
        if is_binary:
            cross_section = _postprocess_cross_section(cross_section, sigma)
        cpr.append(cross_section)
    if num_touching > 0:
        logger.debug(f"Number of border touching cross-sections dropped: {num_touching}")
    cpr = np.stack(cpr, axis=0)
    return cpr


def _compute_tangent_vectors(centerline: np.ndarray) -> np.ndarray:
    """Compute the tangent vectors for a given centerline using a CubicSpline. Assumes that vertices are in natural order
    :param centerline: the centerline as an (N, 3) array
    :return: the tangent vectors for each vertex in the centerline
    """
    # Find the spline points
    s = np.zeros(len(centerline))
    s[1:] = np.cumsum(np.sqrt(np.sum(np.diff(centerline, axis=0) ** 2, axis=1)))
    # Generate the spline
    cs = CubicSpline(s, centerline, bc_type="natural")
    # Compute and normalize the tangents
    tangents = cs(s, 1)
    tangents /= np.linalg.norm(tangents, axis=1)[:, None]
    return tangents


def _compute_orthogonal_vectors(vector: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """Find two orthogonal vectors to the original vector.
    :param vector: the vector to find the orthogonal vectors
    :return: the orthogonal vectors
    """
    if vector[0] == 0 and vector[1] == 0:
        if vector[2] == 0:
            # vec is a zero vector
            return None, None
        # vec is along the z-axis
        return np.array([1, 0, 0]), np.array([0, 1, 0])
    else:
        v1 = np.array([-vector[1], vector[0], 0])
        v1 = v1 / np.linalg.norm(v1)
        v2 = np.cross(vector, v1)
        return v1, v2


def _extract_cross_sectional_slice(arr, point, tangent_vector, slice_size, resolution, is_binary):
    """
    Extract a 2D cross-sectional slice from a 3D array given a point and its tangent vector.

    :param arr: 3D binary NumPy array
    :param point: The point in the array (x, y, z)
    :param tangent_vector: The tangent vector at the point
    :param slice_size: The size of the slice to extract (width, height)
    :param resolution: The resolution of the slice (spacing between points in the grid)
    :param is_binary: Is the array binary or not
    """
    # Find two orthogonal vectors
    v1, v2 = _compute_orthogonal_vectors(tangent_vector)
    if v1 is None or v2 is None:
        raise ValueError("Invalid tangent vector provided.")
    # Generate a meshgrid for the slice
    width, height = slice_size
    x_lin = np.linspace(-width / 2, width / 2, int(width / resolution))
    y_lin = np.linspace(-height / 2, height / 2, int(height / resolution))
    x_grid, y_grid = np.meshgrid(x_lin, y_lin)
    # Map the grid points back to the 3D array indices
    slice_points = point + x_grid[..., np.newaxis] * v1 + y_grid[..., np.newaxis] * v2
    x = np.arange(arr.shape[0])
    y = np.arange(arr.shape[1])
    z = np.arange(arr.shape[2])
    if is_binary:
        # Perform nearest neighbor interpolation in each axis and fill with 0
        method = 'nearest'
        fill_value = 0
    else:
        # Perform linear interpolation in each axis and fill with min value
        method = "linear"
        fill_value = arr.min()
    # Perform the interpolation
    slice_2d = interpn(
        (x, y, z),
        arr,
        slice_points,
        method=method,
        bounds_error=False,
        fill_value=fill_value,
    )
    return slice_2d


def _postprocess_cross_section(cross_section: np.ndarray, sigma: int) -> tuple[np.ndarray, int]:
    """Fill small gaps within the cross-section and apply a gaussian filter to smooth the edges
    :param cross_section: the extracted cross-section from _extract_cross_sectional_slice
    :param sigma: the gaussian kernel sigma
    :return: the smoothed cross-section
    """
    cross_section_labels = [label for label in np.unique(cross_section) if label > 0]
    if len(cross_section_labels) == 0:
        return cross_section
    output_cross_section = np.zeros_like(cross_section, dtype=np.uint8)
    for label in cross_section_labels:
        center_label = _filter_label_regions(cross_section, label)
        center_label = remove_small_holes(center_label)
        center_label = gaussian(center_label, sigma=sigma).round(0)
        # Assign the output of the binary to the appropriate label
        output_cross_section[center_label == 1] = label  
    # # Check if the center label touches the image border
    # height, width = output_cross_section.shape
    # main_label = (output_cross_section == 1).astype(int)
    # if np.any(main_label):
    #     borders = {"top": 0, "bottom": height - 1, "left": 0, "right": width - 1}
    #     touches_border = {
    #         "top": np.any(main_label[borders["top"], :]),
    #         "bottom": np.any(main_label[borders["bottom"], :]),
    #         "left": np.any(main_label[:, borders["left"]]),
    #         "right": np.any(main_label[:, borders["right"]]),
    #     }
    #     touching_any = any(touches_border.values())
    #     if touching_any:
    #         return np.zeros_like(cross_section), 1
    #     main_region = _get_regions(main_label)
    #     solidity = main_region[0].solidity
    #     print(solidity)
    return output_cross_section


def _filter_label_regions(cross_section: np.ndarray, label: int) -> np.ndarray:
    """Filter the labels within a CPR cross-section to only return the label region closest to the center of the image
    :param cross_section: the 2d numpy array containing the entire cross-section
    :param label: the label to filter
    :return: a numpy array containing the filtered cross-section
    """
    tmp_cross_section = cross_section.copy()  # Create a temporary array, so we don't change the original cross-section
    tmp_cross_section[tmp_cross_section != label] = 0  # set anything that is not the desired label to 0
    tmp_cross_section[tmp_cross_section == label] = 1  # set the label to 1 for easier replacement later
    regions = _get_regions(tmp_cross_section)
    if len(regions) == 0:
        centered_label = np.zeros_like(cross_section)
    elif len(regions) > 1:
        centered_label = _closest_to_centroid(tmp_cross_section, regions)
    else:
        centered_label = tmp_cross_section
    return centered_label > 0


def _closest_to_centroid(cross_section: np.ndarray, regions: list[RegionProperties]) -> np.ndarray:
    """Filter a cross-section label using skimage.measure.regionprops to the region closest to the center
    :param cross_section: the cross-section array
    :param regions: the list output from skimage.measure.regionprops
    :return: the filtered numpy array
    """
    center_of_plane = np.array(cross_section.shape) / 2.0
    centroids = [np.array(region.centroid) for region in regions]
    distance_per_region = np.asarray([np.linalg.norm(centroid - center_of_plane) for centroid in centroids])
    min_distance_region_idx = int(np.argmin(distance_per_region))
    center_region = regions[min_distance_region_idx]
    center_label = np.zeros_like(cross_section)
    center_label[center_region.coords[:, 0], center_region.coords[:, 1]] = 1
    return center_label


######### CPR Measurement Functions #############
def measure_largest_cpr_diameter(cpr: np.ndarray, pixel_spacing: tuple, diff_threshold: int = 5) -> dict:
    """Find the largest average cpr diameter
    :param cpr: a binary array containing the entire curved planar reformation to be measured
    :param pixel_spacing: the x, y pixel spacing for each cross-section
    :param diff_threshold: the maximum difference allowed between major and minor diameters for a cross-section.
        If |major - minor| > diff_threshold, set the average diameter = minor diameter to be conservative. Default = 5
    :return: a dictionary with {'avg_diam', 'major_diam', 'minor_diam'} as keys,
        holding the data from the largest diameter cross-section
    """
    avg_diams = []
    major_diams = []
    minor_diams = []
    solidity_threshold = 0.95
    circularity_threshold = 0.9
    eccentricity_threshold = 0.8
    for cross_section in cpr:
        data = measure_cross_sectional_diameter(
            cross_section, pixel_spacing, diff_threshold
        )
        is_convex = data['solidity'] >= solidity_threshold
        is_circular = data['circularity'] >= circularity_threshold and data['circularity'] <= 1.1
        is_not_elliptical = data['eccentricity'] <= eccentricity_threshold
        if is_convex and is_circular and is_not_elliptical:
            avg_diam = data['max_diam']
            major_diam = data['major_diam']
            minor_diam = data['minor_diam']
            avg_diams.append(avg_diam)
            major_diams.append(major_diam)
            minor_diams.append(minor_diam)
    if avg_diams:
        largest_idx = np.argmax(avg_diams)
        diam_dict = {
            "avg_diam": avg_diams[largest_idx],
            "major_diam": major_diams[largest_idx],
            "minor_diam": minor_diams[largest_idx],
        }
    else:
        diam_dict = {}
    return diam_dict


def measure_mid_cpr_diameter(cpr: np.ndarray, pixel_spacing: tuple) -> dict:
    """Measure the mid-region diameter of a curved planar reformation
    :param cpr: a binary array containing the entire curved planar reformation to be measured
    :param pixel_spacing: the x, y pixel spacing for each cross-section
    :return: a dictionary with the mid-region average diameter"""
    len_cpr = len(cpr)
    midslice = len_cpr // 2
    if len_cpr > 0:
        data = measure_cross_sectional_diameter(cpr[midslice], pixel_spacing, diff_threshold=5)
        avg_diam = data["max_diam"]
    else:
        avg_diam = None
    return {"mid_diam": avg_diam}


def measure_cross_sectional_diameter(
    cross_section: np.ndarray, pixel_spacing: tuple, diff_threshold: int
) -> dict[str, float]:
    """Measure the cross-sectional diameter from a straightened cpr slice
    :param cross_section: the binary straightened cpr slice as a numpy array
    :param pixel_spacing: the pixel spacing of the cpr slice
    :param diff_threshold: the maximum difference allowed between major and minor diameters.
        If |major - minor| > diff_threshold, set the average diameter = minor diameter to be conservative
    :return: a dictionary containing the max, major, minor diameters, as well as the solidarity, cicularity, and eccentricity
    """
    regions = _get_regions(cross_section)
    data = {
        'max_diam': 0,
        'major_diam': 0,
        'minor_diam': 0,
        'solidity': 0,
        'circularity': 0,
        'eccentricity': 1.0
    }
    if len(regions) == 0:
        return data
    region = regions[0]
    major_endpoints, minor_endpoints = _get_cross_section_endpoints(region)
    major_units = list(np.multiply(major_endpoints, pixel_spacing))
    minor_units = list(np.multiply(minor_endpoints, pixel_spacing))
    major_diam = _endpoint_euclidean_distance(major_units)
    minor_diam = _endpoint_euclidean_distance(minor_units)
    if abs(major_diam - minor_diam) < diff_threshold:
        max_diam = round((major_diam + minor_diam) / 2, 1)
    else:
        max_diam = min(major_diam, minor_diam)
    solidity = region.solidity 
    circularity = (4 * np.pi * region.area) / (region.perimeter ** 2) if region.perimeter > 0 else 0
    eccentricity = region.eccentricity
    data.update({
        'max_diam': max_diam,
        'major_diam': major_diam,
        'minor_diam': minor_diam,
        'solidity': round(solidity, 2),
        'circularity': round(circularity, 2),
        'eccentricity': round(eccentricity, 2)
    })
    return data


def _get_regions(image: np.ndarray) -> list[RegionProperties]:
    """Get the regions of an image using skimage.measure.regionprops
    :param image: the input image
    :return: the regions of an image
    """
    labels = measure.label(image)
    regions = measure.regionprops(labels)
    return regions


def _get_cross_section_endpoints(
    region: RegionProperties,
) -> tuple[tuple[tuple, tuple], tuple[tuple, tuple]]:
    centroid = region.centroid
    orientation = region.orientation
    major_endpoints = _get_axis_endpoints(centroid, orientation, region.axis_major_length)
    minor_endpoints = _get_axis_endpoints(centroid, orientation, region.axis_minor_length)
    return major_endpoints, minor_endpoints


def _get_axis_endpoints(centroid: np.array, orientation: float, axis_length: float) -> tuple[tuple, tuple]:
    """Calculate the endpoints of the axis of a cross-section region
    :param centroid: the region.centroid
    :param orientation: the region.orientation
    :param axis_length: region.axis_major_length or region.axis_minor_length
    :return: a tuple containing the endpoints
    """
    y0, x0 = centroid
    # calculate the endpoints of the major axis using the centroid
    x1 = x0 - np.sin(orientation) * 0.5 * axis_length
    x2 = x0 + np.sin(orientation) * 0.5 * axis_length
    y1 = y0 - np.cos(orientation) * 0.5 * axis_length
    y2 = y0 + np.cos(orientation) * 0.5 * axis_length
    return (y1, x1), (y2, x2)


def _endpoint_euclidean_distance(endpoints: list) -> float:
    """Calculate the Euclidean distance between endpoints
    :param endpoints: the endpoints (must be two of them)
    :return: the Euclidean distance
    """
    p0, p1 = [*endpoints]
    return round(np.sqrt(((p0 - p1) ** 2).sum()), 1)
